import os
import sys
import signal
import argparse
import _fastpysgi

__version__ = '0.2'

LL_DISABLED    = 0
LL_FATAL_ERROR = 1
LL_CRIT_ERROR  = 2
LL_ERROR       = 3
LL_WARNING     = 4
LL_NOTICE      = 5
LL_INFO        = 6
LL_DEBUG       = 7
LL_TRACE       = 8

class _Server():
    def __init__(self):
        self.app = None
        self.host = "0.0.0.0"
        self.port = 5000
        self.backlog = 2048
        self.loglevel = LL_ERROR
        self.hook_sigint = 2            # 0 = ignore Ctrl-C; 1 = stop server on Ctrl-C; 2 = halt process on Ctrl-C
        self.allow_keepalive = True
        self.add_header_date = True
        self.add_header_server = "FastPySGI/{}".format(__version__)
        self.max_content_length = None  # def value: 999999999
        self.max_chunk_size = None      # def value: 256 KiB
        self.read_buffer_size = None    # def value: 64 KiB
        self.tcp_nodelay = 0            # 0 = Nagle's algo enabled; 1 = Nagle's algo disabled;
        self.tcp_keepalive = 0          # -1 = disabled; 0 = system default; 1...N = timeout in seconds
        self.tcp_send_buf_size = 0      # 0 = system default; 1...N = size in bytes
        self.tcp_recv_buf_size = 0      # 0 = system default; 1...N = size in bytes
        self.nowait = 0
        self.num_workers = 1
        self.worker_list = [ ]
        
    def init(self, app, host = None, port = None, loglevel = None, workers = None):
        self.app = app
        self.host = host if host else self.host
        self.port = port if port else self.port
        self.loglevel = loglevel if loglevel is not None else self.loglevel
        self.num_workers = workers if workers is not None else self.num_workers
        if self.num_workers > 1:
            return 0
        return _fastpysgi.init_server(self)

    def set_allow_keepalive(self, value):
        self.allow_keepalive = value
        _fastpysgi.change_setting(self, "allow_keepalive")

    def run(self):
        if self.nowait:
            if self.num_workers > 1:
                raise Exception('Incorrect server options')
            return _fastpysgi.run_nowait(self)
        if self.num_workers > 1:
            return self.multi_run()
        ret = _fastpysgi.run_server(self)
        self.close()
        return ret
        
    def close(self):
        return _fastpysgi.close_server(self)

    def multi_run(self, num_workers = None):
        if num_workers is not None:
            self.num_workers = num_workers
        for _ in range(self.num_workers):
            pid = os.fork()
            if pid > 0:
                self.worker_list.append(pid)
                print(f"Worker process added with PID: {pid}")
                continue
            try:
                _fastpysgi.init_server(self)
                _fastpysgi.run_server(self)
            except KeyboardInterrupt:
                pass
            sys.exit(0)
        try:
            for _ in range(self.num_workers):
                os.wait()
        except KeyboardInterrupt:
            print("\n" + "Stopping all workers")
            for worker in self.worker_list:
                os.kill(worker, signal.SIGINT)
        return 0

server = _Server()

# -------------------------------------------------------------------------------------

def import_from_string(import_str):
    import importlib
    import importlib.util
    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        raise ImportError("Import string should be in the format <module>:<attribute>")

    try:
        relpath = f"{module_str}.py"
        if os.path.isfile(relpath):
            spec = importlib.util.spec_from_file_location(module_str, relpath)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

        else:
            module = importlib.import_module(module_str)

        for attr_str in attrs_str.split("."):
            module = getattr(module, attr_str)
    except AttributeError:
        raise ImportError(f'Attribute "{attrs_str}" not found in module "{module_str}"')

    return module

# -------------------------------------------------------------------------------------

def run_from_cli():
    parser = argparse.ArgumentParser(
        prog="fastpysgi",
        description="Run FastPySGI server from CLI",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "app_import_string",
        metavar="APP",
        type=str,
        help="WSGI/ASGI app in the format module:attribute (e.g. myapp:app)",
    )
    parser.add_argument(
        "--host",
        type=str,
        default=server.host,
        help=f"Host the socket is bound to (default: {server.host})",
    )
    parser.add_argument(
        "-p", "--port",
        type=int,
        default=server.port,
        help=f"Port the socket is bound to (default: {server.port})",
    )
    parser.add_argument(
        "-l", "--loglevel",
        type=int,
        default=server.loglevel,
        help=f"Logging level 0..8 (default: {server.loglevel})",
    )

    args = parser.parse_args()

    try:
        app = import_from_string(args.app_import_string)
    except ImportError as e:
        print(f"Error importing WSGI app: {e}", file=sys.stderr)
        sys.exit(1)

    server.init(app, args.host, args.port, args.loglevel)
    print(f"FastPySGI server listening at http://{server.host}:{server.port}")
    server.run()

# -------------------------------------------------------------------------------------

def run(app = None, host = None, port = None, loglevel = None, workers = None, wsgi_app = None):
    if app and wsgi_app:
        raise Exception("It is not allowed to specify several applications at once.")
    if app is None:
        app = wsgi_app
    if app is None:
        raise Exception("app not specify.")
    print("FastPySGI server running on PID:", os.getpid())
    server.init(app, host, port, loglevel, workers)
    addon = " multiple workers" if server.num_workers > 1 else ""
    print(f"FastPySGI server{addon} listening at http://{server.host}:{server.port}")
    server.run()



import requests


def test_no_response(general_test_server):
    url = f"{general_test_server.endpoint}/no_response"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text is ''

def test_invalid_return_type(general_test_server):
    url = f"{general_test_server.endpoint}/invalid_return_type"
    result = requests.get(url)
    assert result.status_code == 500



import requests

def test_no_start_response_args(start_response_server):
    url = f"{start_response_server.endpoint}/no_args"
    result = requests.get(url)
    assert result.status_code == 500

def test_invalid_status(start_response_server):
    url = f"{start_response_server.endpoint}/invalid_status"
    result = requests.get(url)
    assert result.status_code == 500

def test_valid_headers(start_response_server):
    url = f"{start_response_server.endpoint}/valid_headers"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text == "OK"    

def test_empty_headers(start_response_server):
    url = f"{start_response_server.endpoint}/empty_headers"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text == "OK"

def test_no_headers(start_response_server):
    url = f"{start_response_server.endpoint}/no_headers"
    result = requests.get(url)
    assert result.status_code == 500
    
def test_wrong_header_type(start_response_server):
    url = f"{start_response_server.endpoint}/wrong_headers"
    result = requests.get(url)
    assert result.status_code == 500
    
def test_wrong_header_value_types(start_response_server):
    url = f"{start_response_server.endpoint}/wrong_header_values"
    result = requests.get(url)
    assert result.status_code == 500

def test_exc_info(start_response_server):
    url = f"{start_response_server.endpoint}/wrong_exc_info_type"
    result = requests.get(url)
    assert result.status_code == 500
    







import os
import sys
import glob

SOURCES = glob.glob('libuv/src/*.c')

if sys.platform == 'win32':
    SOURCES += glob.glob("libuv/src/win/*.c")
else:
    SOURCES += [
        'libuv/src/unix/async.c',
        'libuv/src/unix/core.c',
        'libuv/src/unix/dl.c',
        'libuv/src/unix/fs.c',
        'libuv/src/unix/getaddrinfo.c',
        'libuv/src/unix/getnameinfo.c',
        'libuv/src/unix/loop-watcher.c',
        'libuv/src/unix/loop.c',
        'libuv/src/unix/pipe.c',
        'libuv/src/unix/poll.c',
        'libuv/src/unix/process.c',
        'libuv/src/unix/signal.c',
        'libuv/src/unix/stream.c',
        'libuv/src/unix/tcp.c',
        'libuv/src/unix/thread.c',
        'libuv/src/unix/tty.c',
        'libuv/src/unix/udp.c',
        'libuv/src/unix/random-getrandom.c',
        'libuv/src/unix/random-devurandom.c',
    ]

if sys.platform.startswith('linux'):
    SOURCES += [
        'libuv/src/unix/linux-core.c',
        'libuv/src/unix/linux-inotify.c',
        'libuv/src/unix/linux-syscalls.c',
        'libuv/src/unix/procfs-exepath.c',
        'libuv/src/unix/proctitle.c',
        'libuv/src/unix/random-sysctl-linux.c',
        'libuv/src/unix/epoll.c',
    ]
elif sys.platform == 'darwin':
    SOURCES += [
        'libuv/src/unix/bsd-ifaddrs.c',
        'libuv/src/unix/darwin.c',
        'libuv/src/unix/darwin-proctitle.c',
        'libuv/src/unix/fsevents.c',
        'libuv/src/unix/kqueue.c',
        'libuv/src/unix/proctitle.c',
        'libuv/src/unix/pthread-fixes.c',
        'libuv/src/unix/random-getentropy.c',
    ]
elif sys.platform.startswith(('freebsd', 'dragonfly')):
    SOURCES += [
        'libuv/src/unix/bsd-ifaddrs.c',
        'libuv/src/unix/freebsd.c',
        'libuv/src/unix/kqueue.c',
        'libuv/src/unix/posix-hrtime.c',
    ]
elif sys.platform.startswith('openbsd'):
    SOURCES += [
        'libuv/src/unix/bsd-ifaddrs.c',
        'libuv/src/unix/kqueue.c',
        'libuv/src/unix/openbsd.c',
        'libuv/src/unix/posix-hrtime.c',
    ]
elif sys.platform.startswith('netbsd'):
    SOURCES += [
        'libuv/src/unix/bsd-ifaddrs.c',
        'libuv/src/unix/kqueue.c',
        'libuv/src/unix/netbsd.c',
        'libuv/src/unix/posix-hrtime.c',
    ]

elif sys.platform.startswith('sunos'):
    SOURCES += [
        'libuv/src/unix/no-proctitle.c',
        'libuv/src/unix/sunos.c',
    ]


def build_libuv(build_ext):
    self = build_ext
    self.libuv_dir = os.path.join('libuv')
    if True:
        self.compiler.add_include_dir(os.path.join(self.libuv_dir, 'include'))
        self.compiler.add_include_dir(os.path.join(self.libuv_dir, 'src'))
        self.extensions[0].sources += SOURCES

        if sys.platform != 'win32':
            self.compiler.define_macro('_LARGEFILE_SOURCE', 1)
            self.compiler.define_macro('_FILE_OFFSET_BITS', 64)

        if sys.platform.startswith('linux'):
            self.compiler.define_macro('_GNU_SOURCE', 1)
            self.compiler.add_library('dl')
            self.compiler.add_library('rt')
        elif sys.platform == 'darwin':
            self.compiler.define_macro('_DARWIN_USE_64_BIT_INODE', 1)
            self.compiler.define_macro('_DARWIN_UNLIMITED_SELECT', 1)
        elif sys.platform.startswith('netbsd'):
            self.compiler.add_library('kvm')
        elif sys.platform.startswith('sunos'):
            self.compiler.define_macro('__EXTENSIONS__', 1)
            self.compiler.define_macro('_XOPEN_SOURCE', 500)
            self.compiler.add_library('kstat')
            self.compiler.add_library('nsl')
            self.compiler.add_library('sendfile')
            self.compiler.add_library('socket')
        elif sys.platform == 'win32':
            self.compiler.define_macro('WIN32', 1)
            self.compiler.define_macro('_CRT_SECURE_NO_DEPRECATE', 1)
            self.compiler.define_macro('_CRT_NONSTDC_NO_DEPRECATE', 1)
            self.compiler.define_macro('_WIN32_WINNT', '0x0600')
            self.compiler.add_library('advapi32')
            self.compiler.add_library('iphlpapi')
            self.compiler.add_library('psapi')
            self.compiler.add_library('shell32')
            self.compiler.add_library('user32')
            self.compiler.add_library('userenv')
            self.compiler.add_library('ws2_32')
            self.compiler.add_library('secur32')



import socket
import pytest

BAD_REQUEST_RESPONSE = b"HTTP/1.1 400 Bad Request\r\n\r\n"

raw_requests = [
    "GET\r\n\r\n",
    "/\r\n\r\n",
    "GET ???\r\n\r\n",
    "GET / HTTP\r\n\r\n",
]


@pytest.mark.parametrize("raw_request", raw_requests)
def test_bad_requests(basic_test_server, raw_request):
    host, port = basic_test_server.host, basic_test_server.port
    connection = socket.create_connection((host, port))
    connection.send(raw_request.encode())
    data = connection.recv(4096)
    assert data == BAD_REQUEST_RESPONSE


import requests


def test_wsgi_get(wsgi_test_server):
    url = f"{wsgi_test_server.endpoint}/get"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text == "OK"

def test_wsgi_get_bytestring(wsgi_test_server):
    url = f"{wsgi_test_server.endpoint}/get_byte_string"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text == "basic byte string"


def test_wsgi_post(wsgi_test_server):
    url = f"{wsgi_test_server.endpoint}/post"
    result = requests.post(url, json={"test": "data"})
    assert result.status_code == 201
    assert result.text == "OK"


def test_wsgi_delete(wsgi_test_server):
    url = f"{wsgi_test_server.endpoint}/delete"
    result = requests.delete(url)
    assert result.status_code == 204
    assert result.text == ""



import requests


def test_flask_get(flask_test_server):
    url = f"{flask_test_server.endpoint}/get"
    result = requests.get(url)
    assert result.status_code == 200
    assert result.text == "get"


def test_flask_post(flask_test_server):
    url = f"{flask_test_server.endpoint}/post"
    result = requests.post(url, json={"test": "data"})
    assert result.status_code == 201
    assert result.text == "post"


def test_flask_delete(flask_test_server):
    url = f"{flask_test_server.endpoint}/delete"
    result = requests.delete(url)
    assert result.status_code == 204
    assert result.text == ""



import os
import sys
import subprocess

from fastpysgi import __version__ as fastpysgi_version


def call_fastwsgi(parameters = None, arguments = None, envs = None):
    arguments = arguments or []
    parameters = parameters or []
    envs = envs or {}
    env = os.environ.copy()
    env.update(envs)

    proc = subprocess.run(
        ["fastpysgi"] + arguments + parameters,
        capture_output=True,
        text=True,
        env=env,
    )

    class Result:
        exit_code = proc.returncode
        output = proc.stdout + proc.stderr

    return Result()


class TestCLI:
    def test_help(self):
        result = call_fastwsgi(parameters=["--help"])
        assert result.exit_code == 0
        assert "Run FastPySGI server from CLI" in result.output

    def test_version(self):
        result = call_fastwsgi(parameters=["--version"])
        assert result.exit_code == 0
        assert result.output.strip() == fastpysgi_version

    def test_run_from_cli_invalid_module(self):
        result = call_fastwsgi(arguments=["module:wrong"])
        assert result.exit_code == 1
        assert "Error importing WSGI app: No module named 'module'" in result.output



import os
import sys
import glob
import subprocess
import platform
import re
from setuptools import setup
from distutils.core import Extension
from distutils.command.build_ext import build_ext
import importlib.util

setup_libuv_py = os.path.join(os.path.dirname(__file__), "setup_libuv.py")
spec = importlib.util.spec_from_file_location("setup_libuv", setup_libuv_py)
setup_libuv = importlib.util.module_from_spec(spec)
spec.loader.exec_module(setup_libuv)

SOURCES = glob.glob("fastwsgi/*.c") + glob.glob("llhttp/src/*.c")

module = Extension(
    "_fastpysgi",
    sources=SOURCES,
    include_dirs=["llhttp/include", "libuv/include"],
)

def get_version():
    main_py = os.path.join(os.path.dirname(__file__), "fastpysgi.py")
    with open(main_py, "r", encoding = "utf-8") as file:
        content = file.read()
    match = re.search(r'^__version__\s*=\s*["\']([^"\']+)["\']', content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Cannot find __version__ in fastpysgi.py")

def get_compiler_version(exe_name):
    version = None
    ver_major = None
    ver_minor = None
    try:
        res = subprocess.run([exe_name, '--version'], stdout=subprocess.PIPE)
        if res.returncode == 0:
            version = res.stdout.decode('utf8')
            ver_s = re.search(r"[\D+](\d+)[.](\d+)[\D+]", version)
            if ver_s:
                ver_major = int(ver_s.group(1))
                ver_minor = int(ver_s.group(2))
    except:
        pass
    return version, ver_major, ver_minor


current_compiler = os.getenv('CC', "")

# Forced use clang compiler, if installed
if platform.system() == "Linux" and current_compiler == "":
    for cc_ver in range(40, 7, -1):
        cc = 'clang-%d' % cc_ver
        text, ver_major, ver_minor = get_compiler_version(cc)
        if text and ver_major:
            current_compiler = cc
            print('Change current compiler to {}'.format(cc))
            os.environ['CC'] = cc
            break


class build_all(build_ext):
    def initialize_options(self):
        build_ext.initialize_options(self)

    def build_extensions(self):
        global module

        setup_libuv.build_libuv(self)

        compiler_type = self.compiler.compiler_type
        print("Current compiler type:", compiler_type)
        try:
            compiler_cmd = self.compiler.compiler
            print('Current compiler: {}'.format(compiler_cmd))
        except:
            compiler_cmd = None
        
        compiler = ""
        if compiler_cmd:
            compiler = compiler_cmd[0]
        if compiler_type == "msvc":
            compiler = "msvc"
        
        compiler_ver_major = 0
        if compiler_type == 'unix' and compiler:
            version, c_ver_major, c_ver_minor = get_compiler_version(compiler)
            print('Current compiler version: {}.{}'.format(c_ver_major, c_ver_minor))
            if c_ver_major:
                compiler_ver_major = c_ver_major

        so_file = None
        for ext in self.extensions:
            if ext == module:
                if compiler_type == 'msvc':
                    ext.extra_compile_args += [ '/Oi', '/Oy-', '/W3', '/WX-', '/Gd', '/GS' ]
                    ext.extra_compile_args += [ '/Zc:forScope', '/Zc:inline', '/fp:precise', '/analyze-' ]
                    ext.extra_compile_args += ["/wd4702"]  # disable unreachable code warning
                else:
                    ext.extra_compile_args += [ "-O3", "-fno-strict-aliasing", "-fcommon", "-g", "-Wall" ]
                    ext.extra_compile_args += [ "-Wno-unused-function", "-Wno-unused-variable" ]
                    ext.extra_compile_args += [ "-Wno-unreachable-code" ]  # disable unreachable code warning
                    ext.extra_compile_args += [ "-ffunction-sections", "-fdata-sections" ]
                    if compiler.startswith("gcc") and compiler_ver_major >= 8:
                        ext.extra_compile_args += [ "-Wno-unused-but-set-variable" ]
                    if compiler.startswith("clang") and compiler_ver_major >= 13:
                        ext.extra_compile_args += [ "-Wno-unused-but-set-variable" ]
                    if compiler.startswith("gcc") and sys.platform.startswith("linux"):
                        ext.extra_link_args += [ "-Wl,--gc-sections" ]
                        ext.extra_link_args += [ "-Wl,-O1" ]
                    so_file = self.get_ext_fullpath(ext.name)
        
        build_ext.build_extensions(self)

        if so_file and sys.platform.startswith("linux"):
            print(f"Stripping {so_file}")
            subprocess.run( [ "strip", "--strip-unneeded", so_file ], check = True )


setup(
    version = get_version(),
    ext_modules = [ module ],
    cmdclass = { "build_ext": build_all },
)
